from abc import ABC
from abc import abstractmethod


class Step(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def process(self, data, inputs, utils):
        pass

class StepException(Exception):  # 預防程式錯誤的停止後 須執行的動作
    pass



import sys
sys.path.append('../')
import getopt
import logging
from yt_concate.pipeline.steps.preflight import Preflight
from yt_concate.pipeline.steps.get_video_list import GetVideoList
from yt_concate.pipeline.steps.initialize_yt import InitializeYT
from yt_concate.pipeline.steps.download_captions import DownloadCaptions
from yt_concate.pipeline.steps.read_caption import ReadCaption
from yt_concate.pipeline.steps.search import Search
from yt_concate.pipeline.steps.download_videos import DownloadVideos
from yt_concate.pipeline.steps.edit_video import EditVideo
from yt_concate.pipeline.steps.postflight import Postflight
from yt_concate.pipeline.steps.step import StepException
from yt_concate.pipeline.pipeline import Pipeline
from yt_concate.utils import Utils


def config_logger(loglevel):
    # create logger
    logger = logging.getLogger('mainModule')
    logger.setLevel(logging.DEBUG)

    # create formatter
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    # file handler
    # create file handler and set level to debug
    file_handler = logging.FileHandler('run_information.log')
    file_handler.setLevel(logging.DEBUG)
    # add formatter to file_handler
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)

    # stream handler
    # create stream handler and set level to WARNING (OR user decide)
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(loglevel)
    # add formatter to stream_handler
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    return logger  # 可以不用


def print_usage():
    print('python main.py -c <channel_id> -s <search_word> -l <int(limit)>')
    print('python main.py'
          '--channel_id <channel_id>'
          '--search_word <word>'
          '--limit <number>'
          '--cleanup <True/False>'
          '--fast <True/False>'
          '--log <DEBUG/INFO/WARNING/ERROR/CRITICAL>'
          )

    print('python3 main.py OPTIONS')
    print('OPTIONS: ')
    print('{:>6} {:<12}{}'.format('', '--cleanup', 'Remove captions and video dowloaded during run.'))
    print('{:>6} {:<12}{}'.format('', '--fast', 'skip downloaded captions and videos.'))
    print('{:>6} {:<12}{}'.format('', '--log', 'set the level of logger showed on stream.'))


def main():
    CHANNEL_ID = 'UCKSVUHI9rbbkXhvAXK-2uxA'  # 頻道 id
    WORD = 'incredible'  # 要搜尋的字/詞
    LIMIT = 4  # 合併影片的最高片段數量
    loglevel = logging.WARNING

    inputs = {
        'channel_id': CHANNEL_ID,
        'search_word': WORD,
        'limit': LIMIT,
        'cleanup': False,  # 結果檔產生後，刪除程式執行中產生的檔案，如下載的影片/字幕等
        'fast': True,  # 跳過重複下載
        'loglevel': loglevel
    }

    short_opts = 'hc:s:l:'
    long_opts = 'help channel_id= search_word= limit= cleanup= fast= log='.split()

    try:
        opts, args = getopt.getopt(sys.argv[1:], short_opts, long_opts)  # 用 opts 解析 argv 所投入的參數 ，投遞只能印出檔名後的東西
    except getopt.GetoptError:
        print_usage()
        sys.exit(2)

    for opt, arg in opts:
        if opt == '-h':
            print_usage()
            sys.exit(0)
        elif opt in ('-c', '--channel_id'):
            inputs['channel_id'] = arg
        elif opt in ('-s', '--search_word'):
            inputs['search_word'] = arg
        elif opt in ('-l', '--limit'):
            inputs['limit'] = int(arg)
        elif opt in ('--cleanup'):
            inputs['cleanup'] = bool(arg)
        elif opt in ('--fast'):
            inputs['fast'] = bool(arg)
        elif opt in ('--log'):
            inputs['loglevel'] = eval(f'logging.{arg}')

    print('CHANNEL_ID is ', inputs['channel_id'])
    print('WORD is ', inputs['search_word'])
    print('LIMIT is ', inputs['limit'])
    print('cleanup is ', inputs['cleanup'])
    print('fast is ', inputs['fast'])
    print('log level is ', inputs['loglevel'])

    if not CHANNEL_ID or not WORD:
        print_usage()
        sys.exit(2)

    steps = [
        Preflight(),
        GetVideoList(),
        InitializeYT(),
        DownloadCaptions(),
        ReadCaption(),
        Search(),
        DownloadVideos(),
        EditVideo(),
        Postflight(),
    ]

    config_logger(inputs['loglevel'])
    utils = Utils()
    p = Pipeline(steps)
    p.run(inputs, utils)


if __name__ == '__main__':
    main()



from .step import Step
import logging

class Preflight(Step):
    def process(self, data, inputs, utils):
        logger = logging.getLogger(f'mainModule.{__name__}')
        logger.info('in Preflight')
        utils.create_dirs()


import sys
sys.path.append('C:\\Users\\Meng-Hsin\\venv\\lib\\site-packages')
import logging
import concurrent.futures
import time
from pytube import YouTube
from .step import Step
from .step import StepException
# 多進程變得像一起下載同個東西(好像沒有比較快)

class DownloadCaptions(Step):
    def process(self, data, inputs, utils):
        logger = logging.getLogger(f'mainModule.{__name__}.process')
        start = time.time()

        YT = []
        for yt in data:
            logger.info(f'downloading caption for {yt.id}')
            if utils.caption_file_exist(yt) and inputs['fast'] == True:
                logger.info('found existing caption file')
                pass
            else:
                YT.append(yt)

        for yt in YT:
            with concurrent.futures.ThreadPoolExecutor(max_workers=6) as executor:
                executor.submit(self.download_cap, yt, inputs, utils)

        end = time.time()
        logger.info(f'took, {end - start} , seconds')

        return data

    def download_cap(self, yt, inputs, utils):
        logger = logging.getLogger(f'mainModule.{__name__}.download_cap')
        try:
            source = YouTube(yt.url)
            en_caption = source.captions.get_by_language_code('a.en')
            en_caption_convert_to_srt = (en_caption.generate_srt_captions())
            text_file = open(yt.get_caption_filepath(), "w", encoding='utf-8')
            text_file.write(en_caption_convert_to_srt)
            text_file.close()
        except AttributeError:
            logger.warning(f'AttributeError when downloading caption for {yt.url}')






import logging
from .step import Step

class Postflight(Step):
    def process(self, data, inputs, utils):
        logger = logging.getLogger(f'mainModule.{__name__}')
        if inputs['cleanup'] == True:
            logger.info('in Postflight')
            if utils.output_video_file_exist(data):
                logger.info('found existing output file')
                utils.remove_dirs()
                utils.create_dirs_final()
            else:
                pass


import os
import logging
from threading import Thread
import time
from .step import Step
from pytube import YouTube
from yt_concate.setting import VIDEOS_DIR
# 多線程有排隊拿票的感覺，下載過程中一個線程發現有檔案存在，就趕快再要求另外一個
# 有明顯感覺效率增加

class DownloadVideos(Step):
    def process(self, data, inputs, utils):
        logger = logging.getLogger(f'mainModule.{__name__}.process')
        start = time.time()
        threads = []
        for i in range(4):
            logger.info(f'registering process %d %{i}')
            threads.append(Thread(target=self.download_yt, args = (data[i::4], inputs, utils)))
            # 必須要用 args=() 來做參數傳遞，否則全部跑完才會跳到第二個process
        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

        end = time.time()
        logger.info(f'took, {end - start}, seconds')

        return data

    def download_yt(self, data, inputs, utils):
        logger = logging.getLogger(f'mainModule.{__name__}.download_yt')
        yt_set = set([found.yt for found in data])
        logger.info(f'videos to download, {len(yt_set)}')

        for yt in yt_set:
            url = yt.url

            if utils.video_file_exist(yt) and inputs['fast'] == True:
                logger.info(f'found existing video file for {url}, skipping')
                continue

            if int(len([name for name in os.listdir(VIDEOS_DIR) if os.path.isfile(os.path.join(VIDEOS_DIR, name))])) > inputs['limit']:
                logger.info("the numbers of videos are up to {}".format(inputs['limit']))
                break
            # 因為檔案下載太慢，設條件限制資料夾內影片數量，停止下載
            # 程式碼來源：https://www.itread01.com/content/1549581703.html
            else:
                print('downloading', url)
                YouTube(url).streams.first().download(output_path=VIDEOS_DIR, filename=yt.id)



import logging
from .step import Step
from moviepy.editor import VideoFileClip
from moviepy.editor import concatenate_videoclips


class EditVideo(Step):
    def process(self, data, inputs, utils):
        logger = logging.getLogger(f'mainModule.{__name__}')
        clips = []
        for found in data:
            logger.info(found.time)
            logger.info(found.yt)
            start, end = self.parse_caption_time(found.time)
            try:
                video = VideoFileClip(found.yt.get_video_filepath()).subclip(start, end)
            except OSError:
                logger.warning('no video')
                continue
            # 因為影片太少會出錯找不到路徑 目前先 skip 掉這個步驟
            clips.append(video)
            if len(clips) >= inputs['limit']:
                break
        output_filepath = utils.get_output_filepath(inputs['channel_id'], inputs['search_word'])
        final_clip = concatenate_videoclips(clips)
        final_clip.write_videofile(output_filepath)

        # closing VideoFileClips
        for video in clips:
            video.close()

        return output_filepath

    def parse_caption_time(self, caption_time):
        start, end = caption_time.split(' --> ')
        return self.parse_time_str(start), self.parse_time_str(end)

    def parse_time_str(self, time_str):
        h, m, s = time_str.split(':')
        s, ms = s.split(',')
        return int(h), int(m), int(s) + int(ms)/1000